/* Written by Germano Caronni and Werner Almesberger */
/* (c) by G. Caronni in '94 */
/* This program is under the GNU Public License Version 2 */

extern void init(char *start, double rng, char *clear, char *cipher);
extern int doit(void);
extern char *getresult(void);

