/* Written by Germano Caronni and Werner Almesberger */
/* (c) by G. Caronni in '94 */
/* This program is under the GNU Public License Version 2 */

#ifdef USE_PIPE

#define CPIPE "/u/caronni/dic2/CPIPE"
#define SPIPE "/u/caronni/dic2/SPIPE"
#define RECV recv
#define SEND send

#define RECVFROM(a,b,c,d,e,f) read(a,b,c)
#define SENDTO(a,b,c,d,e,f) write(a,b,c)

struct sockaddr_in { int dummy; } addr;

static int send,recv;

#else

#define RECV s
#define SEND s

#define RECVFROM(a,b,c,d,e,f) recvfrom(a,b,c,d,e,f)
#define SENDTO(a,b,c,d,e,f) sendto(a,b,c,d,e,f)

static int s;
static struct sockaddr_in addr;
#endif
